# mypy: disable-error-code="attr-defined"
import re
import warnings
from functools import cached_property, wraps
from typing import TYPE_CHECKING, Callable, Dict, List, Literal, Optional, Set, cast

from botocraft.services.abstract import PrimaryBoto3ModelQuerySet

if TYPE_CHECKING:
    from botocraft.services import (
        Cluster,
        DeleteTaskDefinitionsResponse,
        Failure,
        Service,
        ServiceDeploymentBrief,
        Task,
        TaskDefinition,
        TaskManager,
    )


# ---------
# Functions
# ---------


def extract_task_family_and_revision(task_definition_arn: str) -> str:
    """
    Extract the task family and revision from a task definition ARN.

    Args:
        task_definition_arn: The ARN of the task definition.

    Returns:
        The task family and revision in the format ``<family>:<revision>``.

    """
    task_definition_arn_re = r"arn:aws:ecs:[^:]+:[^:]+:task-definition/(?P<family>[^:]+):(?P<revision>[0-9]+)"  # noqa: E501
    match = re.match(task_definition_arn_re, task_definition_arn)
    assert match, (
        f"Could not extract task family and revision from {task_definition_arn}"
    )
    return f"{match.group('family')}:{match.group('revision')}"


# ----------
# Decorators
# ----------


# Service


def ecs_services_only(
    func: Callable[..., list[str]],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps :py:meth:`botocraft.services.ecs.ServiceManager.list` to return a
    :py:class:`PrimaryBoto3ModelQuerySet` of
    :py:class:`botocraft.services.ecs.Service` objects instead of only a list of
    ARNs.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        arns = func(self, *args, **kwargs)
        services = []
        # We have to do this in batches of 10 because the get_many method,
        # which uses the boto3 ``describe_services`` method, only accepts 10 ARNs
        # at a time.
        for i in range(0, len(arns), 10):
            services.extend(
                self.get_many(
                    arns[i : i + 10], cluster=kwargs["cluster"], include=["TAGS"]
                ).results
            )
        return PrimaryBoto3ModelQuerySet(services)

    return wrapper


# Cluster


def ecs_clusters_only(
    func: Callable[..., List[str]],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps :py:meth:`botocraft.services.ecs.ClusterManager.list` to return a list
    of :py:class:`botocraft.services.abstract.PrimaryBoto3ModelQuerySet` of
    :py:class:`botocraft.services.ecs.Cluster` objects instead of only a list of
    ARNs.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        arns = func(self, *args, **kwargs)
        clusters: list[Cluster] = []
        # We have to do this in batches of 100 because the get_many method,
        # which uses the boto3 ``describe_clusters`` method, only accepts 100 ARNs
        # at a time.
        for i in range(0, len(arns), 100):
            qs = self.get_many(clusters=arns[i : i + 100], include=["TAGS"])
            clusters.extend(
                qs.results if isinstance(qs, PrimaryBoto3ModelQuerySet) else qs.clusters  # type: ignore[arg-type]
            )
        return PrimaryBoto3ModelQuerySet(clusters)  # type: ignore[arg-type]

    return wrapper


def ecs_task_definition_include_tags(
    func: Callable[..., Optional["TaskDefinition"]],
) -> Callable[..., Optional["TaskDefinition"]]:
    """
    Decorator to convert a :py:class:`botocraft.services.ecs.TaskDefinition` object
    to a :py:class:`botocraft.services.ecs.TaskDefinition` object with tags included.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> Optional["TaskDefinition"]:
        response = func(self, *args, **kwargs)
        if not response:
            return None
        # If we got a TaskDefinition object, we need to convert it to a
        # TaskDefinition with tags.
        _td = response.taskDefinition
        _td.tags = response.tags
        return cast("TaskDefinition", _td)

    return wrapper


def ecs_task_definitions_only(
    func: Callable[..., List[str]],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Decorator to convert a list of ECS task definition identifiers to a list of
    :py:class:`botocraft.services.ecs.TaskDefinition` objects.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        identifiers = func(self, *args, **kwargs)
        responses = [
            self.get(identifier, include=["TAGS"]) for identifier in identifiers
        ]
        return PrimaryBoto3ModelQuerySet(responses)

    return wrapper


# ContainerInstance


def ecs_container_instances_only(
    func: Callable[..., List[str]],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Decorator to convert a list of ECS container instance arns to a
    :py:class:`botocraft.services.abstract.PrimaryBoto3ModelQuerySet` of
    :py:class:`botocraft.services.ecs.ContainerInstance` objects.
    """

    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        arns = func(self, *args, **kwargs)
        container_instances = []
        for i in range(0, len(arns), 100):
            _instances = self.get_many(
                cluster=kwargs["cluster"], containerInstances=arns[i : i + 100]
            )
            if isinstance(_instances, PrimaryBoto3ModelQuerySet):
                _instances = _instances.results
            # If we got a list of ContainerInstanceBrief objects, we need to convert
            if _instances:
                container_instances.extend(_instances)
        return PrimaryBoto3ModelQuerySet(container_instances)

    return wrapper


def ecs_container_instances_tasks_only(
    func: Callable[..., List[str]],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Decorator to convert a list of ECS container instance arns to a list of
    :py:class:`botocraft.services.ecs.ContainerInstance` objects.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        from botocraft.services.ecs import Task

        arns = func(self, *args, **kwargs)
        tasks: list[Task] = []
        for i in range(0, len(arns), 100):
            tasks.extend(
                cast("TaskManager", Task.objects).get_many(arns[i : i + 100]).results  # type: ignore[arg-type]
            )
        return PrimaryBoto3ModelQuerySet(tasks)  # type: ignore[arg-type]

    return wrapper


def ecs_service_deployments_only(
    func: Callable[..., List["ServiceDeploymentBrief"]],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Decorator to convert a list of service deployment arns to a list of
    :py:class:`botocraft.services.ecs.Deployment` objects.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        from botocraft.services.ecs import (
            ServiceDeployment,
            ServiceDeploymentManager,
        )

        response = func(self, *args, **kwargs)
        if response is None:
            return PrimaryBoto3ModelQuerySet([])
        arns = [
            d.serviceDeploymentArn
            for d in func(self, *args, **kwargs)
            if d.serviceDeploymentArn
        ]
        deployments: list[ServiceDeployment] = []
        for i in range(0, len(arns), 20):
            _deployments = cast(
                "ServiceDeploymentManager", ServiceDeployment.objects
            ).get_many(arns[i : i + 20])
            if isinstance(_deployments, PrimaryBoto3ModelQuerySet):
                _deployments = _deployments.results  # type: ignore[assignment]
            # If we got a list of ServiceDeploymentBrief objects, we need to convert
            if _deployments:
                deployments.extend(_deployments)  # type: ignore[arg-type]
        return PrimaryBoto3ModelQuerySet(deployments)  # type: ignore[arg-type]

    return wrapper


# Task
def ecs_task_populate_taskDefinition(
    func: Callable[..., Optional["Task"]],
) -> Callable[..., Optional["Task"]]:
    """
    Wraps :py:meth:`botocraft.services.ecs.TaskManager.get` to populate the
    :py:attr:`botocraft.services.ecs.Task.taskDefinition` attribute.

    We set the ``taskDefinition`` attribute to the task family and revision in the
    format ``<family>:<revision>``.  ``taskDefinition`` is an extra field that we
    add to the :py:class:`botocraft.services.ecs.Task` object that is not in the
    original botocore shape, but is useful for our purposes.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> Optional["Task"]:
        task = func(self, *args, **kwargs)
        if task:
            task.taskDefinition = extract_task_family_and_revision(
                task.taskDefinitionArn
            )
        return task

    return wrapper


def ecs_task_populate_taskDefinitions(
    func: Callable[..., "PrimaryBoto3ModelQuerySet"],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps :py:meth:`botocraft.services.ecs.TaskManager.get_many` to
    populate the :py:attr:`botocraft.services.ecs.Task.taskDefinition` attribute
    on each task.

    We set the ``taskDefinition`` attribute to the task family and revision in the
    format ``<family>:<revision>``.  ``taskDefinition`` is an extra field that we
    add to the :py:class:`botocraft.services.ecs.Task` object that is not in the
    original botocore shape, but is useful for our purposes.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        tasks = func(self, *args, **kwargs)
        for task in tasks:
            task.taskDefinition = extract_task_family_and_revision(
                task.taskDefinitionArn
            )
        return tasks

    return wrapper


def ecs_tasks_only(
    func: Callable[..., List[str]],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wrap :py:meth:`botocraft.services.ecs.TaskManager.list` to return a list of
    :py:class:`botocraft.services.ecs.Task` objects instead of only a list of
    ARNs.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        arns = func(self, *args, **kwargs)
        tasks = []
        # We have to do this in batches of 100 because the get_many method,
        # which uses the boto3 ``describe_tasks`` method, only accepts 100 ARNs
        # at a time.
        for i in range(0, len(arns), 100):
            tasks.extend(
                self.get_many(cluster=kwargs["cluster"], tasks=arns[i : i + 100])
            )
        return PrimaryBoto3ModelQuerySet(tasks)  # type: ignore[arg-type]

    return wrapper


def ecs_task_definition_delete_all(
    func: Callable[..., "DeleteTaskDefinitionsResponse"],
) -> Callable[..., "DeleteTaskDefinitionsResponse"]:
    """
    Decorator to delete all task definitions.  This is because the
    :py:meth:`botocraft.services.ecs.TaskDefinitionManager.delete` method only
    accepts up to 10 task definitions at a time, so we need to delete them in
    batches of 10 if the user passes in more than 10 task definitions.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "DeleteTaskDefinitionsResponse":
        # delete_task_definitions only accepts up to 10 task definitions at a time
        # So we need to delete them in batches
        from botocraft.services import DeleteTaskDefinitionsResponse

        response: DeleteTaskDefinitionsResponse = DeleteTaskDefinitionsResponse(
            taskDefinitions=[],
            failures=[],
        )
        if len(args[0]) > 10:  # noqa: PLR2004
            for i in range(0, len(args[0]), 10):
                response = func(self, args[0][i : i + 10], **kwargs)
                if response.taskDefinitions:
                    cast("list[TaskDefinition]", response.taskDefinitions).extend(
                        response.taskDefinitions
                    )
                if response.failures:
                    cast("list[Failure]", response.failures).extend(response.failures)  # type: ignore[attr-defined]
        else:
            response = func(self, *args, **kwargs)
        return response

    return wrapper


# Mixins


class ECSServiceModelMixin:
    """
    A mixin for :py:class:`botocraft.services.ecs.Service` that adds
    some additional methods that we can't auto generate.
    """

    @property
    def required_cpu(self) -> int:
        """
        The required CPU for the service in CPU shares.  One full CPU is
        equivalent to 1024 CPU shares.
        """
        cpu: int = 0
        td = self.task_definition  # type: ignore[attr-defined]
        if td.cpu:
            cpu = int(td.cpu)
        else:
            for container in td.containerDefinitions:
                if container.cpu:
                    cpu += container.cpu
        return cpu

    @property
    def required_memory(self) -> int:
        """
        Return the required memory for the service in MiB.
        """
        memory: int = 0
        td = self.task_definition  # type: ignore[attr-defined]
        if td.memory:
            memory = int(td.memory)
        else:
            for container in td.containerDefinitions:
                if container.memory:
                    memory += container.memory
        return int(memory)

    @property
    def container_instances(self) -> "PrimaryBoto3ModelQuerySet":
        """
        Return the :py:class:`botocraft.services.ecs.ContainerInstance` objects which
        are running our tasks for the service.
        """
        return PrimaryBoto3ModelQuerySet(
            [task.container_instance for task in self.tasks]  # type: ignore[attr-defined]
        )

    @property
    def is_stable(self) -> bool:
        """
        Return whether the service is stable or not.
        """
        # this is the same test that the `services_stable` waiter uses
        return len(self.deployments) == 1 and (self.runningCount == self.desiredCount)  # type: ignore[attr-defined]

    def wait_until_stable(self, max_attempts: int = 40, delay: int = 15) -> None:
        """
        Wait until the service is stable.

        Raises:
            botocore.exceptions.WaiterError: if the service is not stable after
                ``max_attempts``, or some other error occurred.

        Keyword Args:
            max_attempts: The maximum number of attempts to make before giving
                up.
            delay: The number of seconds to wait between attempts.

        """
        waiter_config = {}
        if max_attempts:
            waiter_config["maxAttempts"] = max_attempts
        if delay:
            waiter_config["delay"] = delay
        if waiter_config:
            waiter_config["operation"] = "DescribeServices"  # type: ignore[assignment]
        waiter = self.objects.using(self.session).get_waiter(
            "services_stable", WaiterConfig=waiter_config
        )  # type: ignore[attr-defined]
        waiter.wait(cluster=self.clusterArn, services=[self.serviceName])  # type: ignore[attr-defined]

    def scale(
        self,
        desired_count: int,
        wait: bool = False,
    ) -> None:
        """
        Scale the service to the desired count.  If ``wait`` is True, this will
        wait for the service to reach the desired count using the ``services_stable``
        boto3 waiter.

        Args:
            desired_count: The number of tasks to run.

        Keyword Args:
            wait: If True, wait for the service to reach the desired count.

        """
        self.objects.using(self.session).partial_update(  # type: ignore[attr-defined]
            self.serviceName,  # type: ignore[attr-defined]
            cluster=self.clusterArn,  # type: ignore[attr-defined]
            desiredCount=desired_count,
        )
        waiter = self.objects.using(self.session).get_waiter("services_stable")  # type: ignore[attr-defined]
        if wait:
            waiter.wait(
                cluster=self.clusterArn,  # type: ignore[attr-defined]
                services=[self.serviceName],  # type: ignore[attr-defined]
            )

    @property
    def load_balancers(self) -> "PrimaryBoto3ModelQuerySet":
        """
        Return the :py:class:`LoadBalancer` objects that are associated with the
        service.
        """
        from botocraft.services import LoadBalancer

        arns: Set[str] = set()
        for tg in self.target_groups:  # type: ignore[attr-defined]
            for arn in tg.LoadBalancerArns:
                arns.add(arn)
        if arns:
            return LoadBalancer.objects.using(self.session).list(
                LoadBalancerArns=list(arns)
            )
        return PrimaryBoto3ModelQuerySet([])  # type: ignore[arg-type]


class ECSServiceManagerMixin:
    """
    A mixin for :py:class:`botocraft.services.ecs.ServiceManager` that adds
    some additional methods that we can't auto generate.
    """

    def all(
        self,
        launchType: Literal["EC2", "FARGATE", "EXTERNAL"] | None = None,  # noqa: N803
        schedulingStrategy: Literal["REPLICA", "DAEMON"] | None = None,  # noqa: N803
        tags: Dict[str, str] | None = None,
    ) -> "PrimaryBoto3ModelQuerySet":
        """
        Return all the services in the account.  This differs from
        :py:meth:`botocraft.services.ServiceManager.list` in that it iterates
        through all the clusters in the account and gets the services for each
        cluster.

        Normally you would expect to use
        :py:meth:`botocraft.services.ServiceManager.list` to get all the
        services, but ``describe_services``, on which our method is based, only
        returns services for a single cluster, so we need to roll our own
        method.

        Args:
            launchType: The launch type of the services to return.
            schedulingStrategy: The scheduling strategy of the services to return.
            tags: A dictionary of tags to filter the services

        Returns:
            A list of :py:class:`Service` objects.

        """
        from botocraft.services import Cluster, Service

        if not tags:
            tags = {}

        clusters = Cluster.objects.using(self.session).list()
        services: List["Service"] = []  # noqa: UP037
        for cluster in clusters:
            if tags.items() <= cluster.tags.items():
                services.extend(
                    Service.objects.using(self.session).list(
                        cluster=cluster.clusterArn,
                        launchType=launchType,
                        schedulingStrategy=schedulingStrategy,
                    )
                )
        return PrimaryBoto3ModelQuerySet(services)  # type: ignore[arg-type]


class ECSContainerInstanceModelMixin:
    @property
    def free_cpu(self) -> int:
        """
        Return the free CPU shares on the container instance.  One full CPU is
        equivalent to 1024 CPU shares.
        """
        value: int = 0
        for resource in self.remainingResources:  # type: ignore[attr-defined]
            if resource.name == "CPU":
                value = int(resource.integerValue)
        return value

    @property
    def free_ram(self) -> int:
        """
        Return the free RAM in MiB on the container instance.
        """
        value: int = 0
        for resource in self.remainingResources:  # type: ignore[attr-defined]
            if resource.name == "MEMORY":
                value = int(resource.integerValue)
        return value


class TaskDefinitionManagerMixin:
    def in_use(
        self,
        tags: Dict[str, str] | None = None,
    ) -> "PrimaryBoto3ModelQuerySet":
        """
        Return a list of task definitions that are currently in use by a service
        or periodic task.  A periodic task is a task that is run via a
        :py:class:`botocraft.services.events.EventRule`.

        Important:
            If you have tasks that are run ad-hoc, then this method will not
            return those task definitions.

        Keyword Args:
            tags: A dictionary of tags to filter the task, services and periodic
                tasks by.  Default: None
            verbose: If True, print out some information about what is happening.

        Returns:
            A list of :py:class:`botocraft.services.ecs.TaskDefinition` objects that
            are currently in use.

        """
        from botocraft.services import (
            EventRule,
            Service,
            TaskDefinition,
        )

        if not tags:
            tags = {}

        task_definitions: Dict[str, TaskDefinition] = {}

        # First get all the services in the account
        services: List[Service] = Service.objects.using(self.session).all()
        ClientException = self.session.client("ecs").exceptions.ClientException  # noqa: N806

        # Now iterate through each service and get the append its task definition
        # to the list of task definitions if we have not already seen it
        for service in services:
            try:
                task_definition = cast("TaskDefinition", service.task_definition)
            except ClientException:
                warnings.warn(
                    f"Task definition {service.taskDefinition} used by "
                    f"{service.cluster_name}:{service.serviceName} does not exist",
                    UserWarning,
                    stacklevel=2,
                )
                continue
            family_revision = task_definition.family_revision
            if family_revision not in task_definitions:
                task_definitions[family_revision] = task_definition

        # Now deal with the periodc tasks
        rules = EventRule.objects.using(self.session).list()
        for rule in rules:
            for target in rule.targets:
                if target.EcsParameters is not None:
                    try:
                        task_definition = TaskDefinition.objects.using(
                            self.session
                        ).get(target.EcsParameters.TaskDefinitionArn)
                    except ClientException:
                        warnings.warn(
                            f"Task definition {target.EcsParameters.TaskDefinitionArn} "
                            f"used by {rule.name} does not exist",
                            UserWarning,
                            stacklevel=2,
                        )
                        continue
                    family_revision = task_definition.family_revision
                    if family_revision not in task_definitions:
                        task_definitions[family_revision] = task_definition
        return PrimaryBoto3ModelQuerySet(list(task_definitions.values()))  # type: ignore[arg-type]


class TaskDefinitionModelMixin:
    @property
    def family_revision(self) -> str:
        """
        Return the family and revision of the task definition in the format
        ``<family>:<revision>``.
        """
        return f"{self.family}:{self.revision}"

    @property
    def container_images(self) -> List[str]:
        """
        Return the container images as a list of strings.

        Returns:
            A list of container images.

        """
        return [container.image for container in self.containerDefinitions]  # type: ignore[attr-defined]

    @cached_property
    def services(self) -> List["Service"]:
        """
        Return the services that use this task definition revision.

        Warning:
            This will be quite slow because we need to all our services
            to see if there is a service that uses that task definition.  There's
            no way to get all the services in an account, so we have to list
            all the clusters, then check each cluster for services, and see
            if the service uses this task definition.

        Returns:
            A list of :py:class:`botocraft.services.ecs.Service` objects that use
            this task definition.

        """
        from botocraft.services import Cluster, Service

        clusters = Cluster.objects.using(self.session).list()
        services: List[Service] = []
        for cluster in clusters:
            services.extend(
                service
                for service in cluster.services
                if service.taskDefinition == self.family_revision
            )
        return services


class ServiceDeploymentModelMixin:
    """
    A mixin for :py:class:`botocraft.services.ecs.ServiceDeployment` that adds
    some additional methods that we can't auto generate.
    """

    @property
    def source_task_definitions(self) -> "PrimaryBoto3ModelQuerySet":
        """
        Return the task definition for the deployment.
        """
        from botocraft.services import TaskDefinition

        arns = [source.arn for source in self.sourceServiceRevisions]

        return PrimaryBoto3ModelQuerySet(
            [TaskDefinition.objects.using(self.session).get(arn) for arn in arns]
        )  # type: ignore[arg-type]



from __future__ import annotations

from datetime import datetime  # noqa: TC003
from typing import List

from pydantic import BaseModel, Field


class ECRReferrerAction(BaseModel):
    """
    Represents details about an action performed on an ECR referrer.  This
    includes pulls and other operations on container images referenced by an
    upstream image or repository.
    """

    #: The type of action performed on the referrer (e.g., "PULL")
    action_type: str = Field(..., alias="action-type")
    #: The SHA256 digest of the image being referenced
    image_digest: str | None = Field(None, alias="image-digest")
    #: The tag associated with the referenced image, if any
    image_tag: str | None = Field(None, alias="image-tag")
    #: The name of the ECR repository containing the referenced image
    repository_name: str = Field(..., alias="repository-name")
    #: The result of the action (e.g., "SUCCESS")
    result: str
    #: The media type of the image manifest
    manifest_media_type: str | None = Field(None, alias="manifest-media-type")
    #: The media type of the artifact
    artifact_media_type: str | None = Field(None, alias="artifact-media-type")


class ECRReferrerActionEvent(BaseModel):
    """
    Represents a complete ECR Referrer Action event from EventBridge.
    Contains metadata about the event and the actual referrer action details.
    """

    #: The detailed information about the ECR referrer action
    detail: ECRReferrerAction
    #: The AWS account ID where the event occurred
    account: str
    #: The human-readable type of the event (e.g., "ECR Referrer Action")
    detail_type: str = Field(..., alias="detail-type")
    #: The unique identifier for this event instance
    id: str
    #: The AWS region where the event occurred
    region: str
    #: The AWS resources involved in this event, typically the repository ARN
    resources: List[str]
    #: The source of the event (typically "aws.ecr")
    source: str
    #: The timestamp when the event was generated, in UTC
    time: datetime
    #: The version of the event schema format
    version: str



from .aws_api_call_via_cloudtrail import (
    ECSAWSAPICallViaCloudTrailEvent,  # noqa: F401
)
from .container_instance_change import (
    ECSContainerInstanceStateChangeEvent,  # noqa: F401
)
from .service_action import ECSServiceActionEvent  # noqa: F401
from .service_deployment_state_change import (
    ECSServiceDeploymentStateChangeEvent,  # noqa: F401
)
from .task_state_change import ECSTaskStateChangeEvent  # noqa: F401



from __future__ import annotations

from datetime import datetime  # noqa: TC003
from enum import Enum
from typing import List

from pydantic import BaseModel, Field


class ScanFrequency(Enum):
    """
    Enumeration of possible ECR image scan frequency settings.
    Defines how often AWS ECR will scan container images for vulnerabilities.
    """

    #: Scan images only when they are first pushed to the repository
    SCAN_ON_PUSH = "SCAN_ON_PUSH"
    #: Continuously scan images in the repository
    CONTINUOUS_SCAN = "CONTINUOUS_SCAN"
    #: Only scan images when manually requested
    MANUAL = "MANUAL"


class Repository(BaseModel):
    """
    Represents an ECR repository affected by a scan resource change.  Contains
    details about the repository and how its scan configuration has changed.
    """

    #: The name of the ECR repository
    repository_name: str = Field(..., alias="repository-name")
    #: The ARN (Amazon Resource Name) of the ECR repository
    repository_arn: str = Field(..., alias="repository-arn")
    #: The new scan frequency setting for the repository
    scan_frequency: ScanFrequency = Field(..., alias="scan-frequency")
    #: The previous scan frequency setting before this change
    previous_scan_frequency: ScanFrequency = Field(
        ...,
        alias="previous-scan-frequency",
    )


class ScanType(Enum):
    """
    Enumeration of ECR scan types.
    Defines the type of vulnerability scanning being performed on ECR images.
    """

    #: Enhanced scanning uses Amazon Inspector and provides more comprehensive results
    ENHANCED = "ENHANCED"
    #: Basic scanning provides fundamental vulnerability detection
    BASIC = "BASIC"


class ScanResourceChange(BaseModel):
    """
    Represents details about a change to ECR scan resource configuration.
    Contains information about what type of change occurred and which
    repositories were affected.
    """

    #: The type of scan resource change action (e.g., "SCAN_FREQUENCY_CHANGE")
    action_type: str = Field(..., alias="action-type")
    #: List of repositories affected by the scan resource change
    repositories: List[Repository] = Field(default_factory=list)
    #: The type of resource affected by the scan change (typically "REPOSITORY")
    resource_type: str = Field(..., alias="resource-type")
    #: The type of scan configured (ENHANCED or BASIC)
    scan_type: ScanType = Field(..., alias="scan-type")


class ECRScanResourceChangeEvent(BaseModel):
    """
    Represents a complete ECR Scan Resource Change event from EventBridge.
    Contains metadata about the event and the actual scan resource change details.
    """

    #: The detailed information about the scan resource change
    detail: ScanResourceChange
    #: The version of the event schema format
    version: str
    #: The unique identifier for this event instance
    id: str
    #: The human-readable type of the event (e.g., "ECR Scan Resource Change")
    detail_type: str = Field(..., alias="detail-type")
    #: The source of the event (typically "aws.ecr")
    source: str
    #: The AWS account ID where the event occurred
    account: str
    #: The timestamp when the event was generated, in UTC
    time: datetime
    #: The AWS region where the event occurred
    region: str
    #: The AWS resources involved in this event, typically repository ARNs
    resources: List[str]



from ...models import MethodDocstringDefinition
from .create import CreateMethodGenerator


class UpdateMethodGenerator(CreateMethodGenerator):
    method_name: str = "update"

    @property
    def docstrings_def(self) -> MethodDocstringDefinition:
        """
        Return the docstring for the method.
        """
        docstrings_def = super().docstrings_def
        docstrings_def.args["model"] = f"The :py:class:`{self.model_name}` to update."
        return docstrings_def



from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, cast

from .base import EventBridgeEvent
from .raw import (
    ECSAWSAPICallViaCloudTrailEvent as RawECSAWSAPICallViaCloudTrailEvent,
)
from .raw import (
    ECSContainerInstanceStateChangeEvent as RawECSContainerInstanceStateChangeEvent,
)
from .raw import (
    ECSServiceActionEvent as RawECSServiceActionEvent,
)
from .raw import (
    ECSServiceDeploymentStateChangeEvent as RawECSServiceDeploymentStateChangeEvent,
)
from .raw import (
    ECSTaskStateChangeEvent as RawECSTaskStateChangeEvent,
)

if TYPE_CHECKING:
    from botocraft.services.ecs import (
        Cluster,
        ContainerInstance,
        Service,
        Task,
        TaskDefinition,
    )


@dataclass
class SystemResources:
    """
    A dataclass that represents the system resources of a container instance,
    container or task.
    """

    #: The CPU units available on the container instance, container or task.
    cpu: float | None = None
    #: The memory available on the container instance, container or task.
    memory: float | None = None


class ECSTaskStateChangeEvent(EventBridgeEvent, RawECSTaskStateChangeEvent):
    """
    EventBridge event for ECS task state change.
    """

    def __str__(self) -> str:
        """
        Return a string representation of the event.
        """
        cluster_name = self.detail.clusterArn.split("/")[-1]
        family_and_revision = self.detail.taskDefinitionArn.split("/")[-1]
        return (
            f"<Event: ECS Task State Change: [{self.account}]: "
            f"source={self.source}, "
            f"time={self.time}, "
            f"region={self.region}, "
            f"resources={self.resources}, "
            f"task_definition={family_and_revision}, "
            f"cluster={cluster_name}, "
            f"desired_status={self.detail.desiredStatus}>"
        )

    @property
    def task_definition(self) -> "TaskDefinition":
        """
        Get the :py:class:`~botocraft.services.ecs.TaskDefinition` object
        from the event.


        """
        from botocraft.services.ecs import TaskDefinition

        return TaskDefinition.objects.using(self.session).get(
            taskDefinition=self.detail.taskDefinitionArn
        )

    @property
    def cluster(self) -> "Cluster":
        """
        Get the cluster name from the event.
        """
        from botocraft.services.ecs import Cluster

        return Cluster.objects.using(self.session).get(cluster=self.detail.clusterArn)

    @property
    def container_instance(self) -> Optional["ContainerInstance"]:
        """
        Get the container instance ARN from the event.
        """
        from botocraft.services.ecs import ContainerInstance

        if not self.detail.containerInstanceArn:
            return None
        return ContainerInstance.objects.using(self.session).get(
            self.detail.containerInstanceArn, cluster=self.detail.clusterArn
        )

    @property
    def task(self) -> str:
        """
        Get the task ARN from the event.
        """
        from botocraft.services.ecs import Task

        return Task.objects.using(self.session).get(
            task=self.detail.taskArn, cluster=self.detail.clusterArn
        )

    @property
    def architecture(self) -> str:
        """
        Get the architecture of the task from the event.
        """
        for attribute in self.detail.attributes:
            if attribute.name == "ecs.architecture":
                return cast("str", attribute.value)
        return "x86_64"

    @property
    def is_fargate(self) -> bool:
        """
        Check if the task is a Fargate task.
        """
        return self.detail.launchType == "FARGATE"


class ECSServiceActionEvent(EventBridgeEvent, RawECSServiceActionEvent):
    """
    EventBridge event for ECS service action.
    """

    def __str__(self) -> str:
        """
        Return a string representation of the event.
        """
        cluster_name = self.detail.clusterArn.split("/")[-1]
        service_names = [resource.split("/")[-1] for resource in self.resources or []]
        return (
            f"<Event: ECS Service Action: account={self.account}, "
            f"source={self.source}, "
            f"time={self.time}, "
            f"region={self.region}, "
            f"services={service_names}, "
            f"cluster={cluster_name}, "
            f"event_type={self.detail.eventType}, "
            f"event_name={self.detail.eventName}>"
        )

    @property
    def cluster(self) -> "Cluster":
        """
        Get the cluster name from the event.
        """
        from botocraft.services.ecs import Cluster

        return Cluster.objects.get(cluster=self.detail.clusterArn)

    @property
    def tasks(self) -> list["Task"] | None:
        """
        Get the task ARNs from the event.
        """
        from botocraft.services.ecs import Task

        if not self.detail.taskArns:
            return None
        return [
            Task.objects.using(self.session).get(
                task=taskArn, cluster=self.detail.clusterArn
            )
            for taskArn in self.detail.taskArns
        ]

    @property
    def container_instances(self) -> list["ContainerInstance"] | None:
        """
        Get the container instance ARN from the event.
        """
        from botocraft.services.ecs import ContainerInstance

        if not self.detail.containerInstanceArns:
            return None

        return [
            ContainerInstance.objects.using(self.session).get(
                containerInstanceArn=containerInstanceArn,
                cluster=self.detail.clusterArn,
            )
            for containerInstanceArn in self.detail.containerInstanceArns
        ]

    @property
    def services(self) -> list["Service"]:
        """
        Get the service ARN from the event.
        """
        from botocraft.services.ecs import Service

        return [
            Service.objects.using(self.session).get(
                service=service_arn, cluster=self.detail.clusterArn
            )
            for service_arn in self.resources or []
        ]

    @property
    def event_type(self) -> str:
        """
        Get the event type from the event.
        """
        return self.detail.eventType


class ECSContainerInstanceStateChangeEvent(
    EventBridgeEvent, RawECSContainerInstanceStateChangeEvent
):
    """
    EventBridge event for ECS container instance state change.
    """

    def __str__(self) -> str:
        """
        Return a string representation of the event.
        """
        cluster_name = self.detail.clusterArn.split("/")[-1]
        return (
            f"<Event: ECS Container Instance State Change: account={self.account}, "
            f"source={self.source}, "
            f"time={self.time}, "
            f"region={self.region}, "
            f"cluster={cluster_name}, "
            f"ec2_instance={self.detail.ec2InstanceId}, "
            f"status={self.detail.status}, "
            f"status_reason={self.detail.statusReason}>"
        )

    @property
    def cluster(self) -> "Cluster":
        """
        Get the cluster name from the event.
        """
        from botocraft.services.ecs import Cluster

        return Cluster.objects.using(self.session).get(cluster=self.detail.clusterArn)

    @property
    def container_instance(self) -> "ContainerInstance":
        """
        Get the container instance ARN from the event.
        """
        from botocraft.services.ecs import ContainerInstance

        return ContainerInstance.objects.using(self.session).get(
            containerInstance=self.detail.containerInstanceArn,
            cluster=self.detail.clusterArn,
        )

    @property
    def remainingResources(self) -> SystemResources:
        """
        Return the remaining resources (CPU and memory) on the container
        instance.

        """
        for resource in self.detail.remainingResources:
            if resource.name == "CPU":
                cpu = resource.integerValue
            elif resource.name == "MEMORY":
                memory = resource.integerValue
        return SystemResources(cpu=cpu, memory=memory)

    @property
    def registeredResources(self) -> SystemResources:
        """
        Return the full system resources (CPU and memory) of the container
        instance.

        """
        for resource in self.detail.registeredResources:
            if resource.name == "CPU":
                cpu = resource.integerValue
            elif resource.name == "MEMORY":
                memory = resource.integerValue
        return SystemResources(cpu=cpu, memory=memory)


class ECSServiceDeploymentStateChangeEvent(
    EventBridgeEvent, RawECSServiceDeploymentStateChangeEvent
):
    """
    EventBridge event for ECS service deployment state change.
    """

    def __str__(self) -> str:
        """
        Return a string representation of the event.
        """
        return (
            f"<Event: ECS Service Deployment Change: account={self.account}, "
            f"source={self.source}, "
            f"time={self.time}, "
            f"region={self.region}, "
            f"resources={self.resources}, "
            f"event_type={self.detail.eventType}, "
            f"event_name={self.detail.eventName}, "
            f"reason={self.detail.reason}>"
        )

    @property
    def cluster(self) -> "Cluster":
        """
        Get the cluster name from the event.
        """
        from botocraft.services.ecs import Cluster

        return Cluster.objects.using(self.session).get(cluster=self.detail.clusterArn)  # type: ignore[attr-defined]

    @property
    def service(self) -> "Service":
        """
        Get the service ARN from the event.
        """
        from botocraft.services.ecs import Service

        service_arn = self.resources[0]
        service_name = service_arn.split("/")[-1]
        cluster_name = service_arn.split("/")[-2]
        return Service.objects.using(self.session).get(
            service=service_name, cluster=cluster_name
        )


class ECSAWSAPICallViaCloudTrailEvent(
    EventBridgeEvent, RawECSAWSAPICallViaCloudTrailEvent
):
    """
    EventBridge event for ECS API call via CloudTrail.
    """

    def __str__(self) -> str:
        """
        Return a string representation of the event.
        """
        return (
            f"<Event: ECS API Call via CloudTrail: account={self.account}, "
            f"source={self.source}, "
            f"time={self.time}, "
            f"region={self.region}, "
            f"resources={self.resources}, "
            f"event_type={self.detail.eventType}, "
            f"event_name={self.detail.eventName}>"
        )

    @property
    def clusters(self) -> list["Cluster"]:
        """
        Get the cluster name from the event.
        """
        from botocraft.services.ecs import Cluster

        clusters: list[Cluster] = []
        if not self.detail.responseElements:
            return clusters
        if not self.detail.responseElements.tasks:
            return clusters
        for task in self.detail.responseElements.tasks:
            clusters.append(  # noqa: PERF401
                Cluster.objects.using(self.session).get(cluster=task.clusterArn)
            )
        return clusters

    @property
    def task_definitions(self) -> list["TaskDefinition"]:
        """
        Get the task definition name from the event.
        """
        from botocraft.services.ecs import TaskDefinition

        task_definitions: list[TaskDefinition] = []
        if not self.detail.responseElements:
            return task_definitions
        if not self.detail.responseElements.tasks:
            return task_definitions
        for task in self.detail.responseElements.tasks:
            task_definitions.append(  # noqa: PERF401
                TaskDefinition.objects.using(self.session).get(
                    taskDefinition=task.taskDefinitionArn
                )
            )
        return task_definitions

    @property
    def tasks(self) -> list["Task"]:
        """
        Get the task ARN from the event.
        """
        from botocraft.services.ecs import Task

        tasks: list[Task] = []
        if not self.detail.responseElements:
            return tasks
        if not self.detail.responseElements.tasks:
            return tasks
        for task in self.detail.responseElements.tasks:
            tasks.append(  # noqa: PERF401
                Task.objects.using(self.session).get(
                    task=task.taskArn, cluster=task.clusterArn
                )
            )
        return tasks

    @property
    def container_instances(self) -> list["ContainerInstance"]:
        """
        Get the container instance ARN from the event.
        """
        from botocraft.services.ecs import ContainerInstance

        container_instances: list[ContainerInstance] = []
        if not self.detail.responseElements:
            return container_instances
        if not self.detail.responseElements.tasks:
            return container_instances
        for task in self.detail.responseElements.tasks:
            if task.containerInstanceArn:
                container_instances.append(  # noqa: PERF401
                    ContainerInstance.objects.using(self.session).get(
                        containerInstance=task.containerInstanceArn,
                        cluster=task.clusterArn,
                    )
                )
        return container_instances



from functools import cached_property
import re
import inspect
from collections import OrderedDict
from datetime import datetime, time, timezone as dt_timezone
from zoneinfo import ZoneInfo
from typing import (
    Any,
    ClassVar,
    Type,
    Callable,
    Iterator,
    cast,
    Union,
    Optional,
    Dict,
    Set,
)

import boto3
from pydantic import BaseModel, ConfigDict, Field

from .exceptions import NotUpdatableError


class TransformMixin:
    def transform(
        self,
        attribute: str,
        transformer: str | None,
    ) -> Any:
        """
        Transform an attribute using a regular expression into something else
        before it is returned.

        .. important::
            This only makes sense for attributes that are strings.

        ``transformer`` is a regular expression that will be used to transform
        the value of the attribute.

        * If the attribute is ``None``, it will be returned verbatim.
        * If ``transformer`` is ``None``, the attribute will be returned verbatim.
        * If ``transformer`` has no named groups, the attribute will be replaced
          with the value of the first group.
        * If ``transformer`` has named groups, the attribute will be replaced
          with a dictionary of the named groups.

        Raises:
            ValueError: If the attribute does not exist on the model.
            RuntimeError: If the transformer fails to match the attribute value.

        Args:
            attribute: The attribute to transform.
            transformer: The regular expression to use to transform the attribute.

        Returns:
            The transformed attribute.

        """
        if not hasattr(self, attribute):
            msg = f"Invalid attribute: {self.__class__.__name__}.{attribute}"
            raise ValueError(msg)
        if transformer is None:
            return getattr(self, attribute)
        value = getattr(self, attribute)
        if value is None:
            return None
        if match := re.search(transformer, value):
            if match.groupdict():
                return match.groupdict()
            return match.group(1)
        msg = (
            f"Transformer failed to match: transformer=r'{transformer}', "
            f"value='{getattr(self, attribute)}'"
        )
        raise RuntimeError(msg)


class Boto3Model(TransformMixin, BaseModel):
    """
    The base class for all boto3 models.
    """

    model_config = ConfigDict(validate_assignment=True)

    #: The boto3 session to use for this model.  This is set by the manager,
    #: and is used in relationships.  We have to use ``Any`` here because we
    #: pydantic complains vociferously if we use ``boto3.session.Session``.
    #: We exclude it from the model dump because it's not something that should
    #: be serialized.
    session: Any | None = Field(default=None, exclude=True)

    def set_session(self, session: boto3.session.Session) -> None:
        """
        Set the boto3 session for this model.

        Args:
            session: The boto3 session to use.

        Returns:
            The model instance.

        """
        if self.model_config.get("frozen"):
            self.model_config["frozen"] = False
            self.session = session
            self.model_config["frozen"] = True
        else:
            self.session = session


class ReadonlyBoto3Model(Boto3Model):
    """
    The base class for all boto3 models that are readonly.
    """

    model_config = ConfigDict(frozen=True, validate_assignment=True, extra="allow")


class Boto3ModelManager(TransformMixin):
    #: The name of the boto3 service.  Example: ``ec2``, ``s3``, etc.
    service_name: str

    def __init__(self) -> None:
        #: The boto3 client for the AWS service
        self.client = boto3.client(self.service_name)  # type: ignore[call-overload]
        #: The boto3 session to use for this manager.
        self.session = boto3.session.Session()

    def using(self, session: boto3.session.Session | None) -> "Boto3ModelManager":
        """
        Use a different boto3 session for this manager.

        Args:
            session: The boto3 session to use.

        """
        # TODO: this is a bad way to do this -- it means that whatever session
        # was last set with .using() will be transparently used by all other
        # calls to the manager.  This is not good.  We need som way to make
        # it like a context manager, so that the session is only used for the
        # duration of the actual method call.
        if session is not None:
            self.session = session
            self.client = session.client(self.service_name)  # type: ignore[call-overload]
        return self

    def serialize(self, arg: Any) -> Any:
        """
        Some of our botocraft methods use :py:class:`Boto3Model` objects as
        arguments (e.g. ``create``, ``update``), but boto3 methods expect simple
        Python types.  This method will serialize the model into a set of types
        that boto3 will understand if it is a :py:class:`Boto3Model` object.

        While serializing, we always exclude ``None`` values, because boto3
        doesn't like them.

        If ``arg`` is not a :py:class:`Boto3Model` object, it will be returned
        verbatim.

        Args:
            arg: the botocraft method argument to serialize.

        Returns:
            A properly serialized argument.

        """
        if arg is None:
            return None
        if isinstance(arg, Boto3Model):
            return arg.model_dump(exclude_none=True)
        if isinstance(arg, list):
            # Oop, this is a list.  We need to serialize each item in the list.
            return [self.serialize(a) for a in arg]
        return arg

    def sessionize(self, response: Any) -> None:
        """
        Look through ``response`` for any object with ``set_session`` as
        an attribute and set the session on that object.

        .. note::

            I'm making an assumption here that the only PrimaryBoto3Model
            or ReadonlyPrimaryBoto3Model objects that will be in the response
            will be at the top level.  If they are nested, we'll miss the ones
            that are nested.

        Args:
            response: The response to search.
        """
        if response:
            if isinstance(response, BaseModel):
                # Pydantic models are ALWAYS iterable, so we can't use
                # iter(response) to check if it's a list.
                # We get the fields on the model, excluding our special
                # fields and then try to sessionize them.
                if isinstance(response, PrimaryBoto3Model | ReadonlyPrimaryBoto3Model):
                    response.set_session(self.session)
                    return
                attrs = [
                    attr
                    for attr in response.__fields__
                    if attr not in ["session", "objects"]
                ]
                for attr in attrs:
                    _attr = getattr(response, attr)
                    if _attr is not None:
                        self.sessionize(_attr)
            else:
                # This is NOT a pydantic model, now we test for iterability, because
                # it could be a list of pydantic models.
                try:
                    iter(response)
                except TypeError:
                    # This is not iterable, so we can't sessionize it.
                    pass
                else:
                    if not response:
                        return
                    # Test for a dict
                    if isinstance(response, dict):
                        for value in response.values():
                            self.sessionize(value)
                    # This is a list or tuple
                    elif hasattr(response[0], "set_session"):
                        [self.sessionize(obj) or obj for obj in response]  # type: ignore[func-returns-value]

    def get(self, *args, **kwargs):
        raise NotImplementedError

    def list(self, *args, **kwargs):
        raise NotImplementedError

    def create(self, model, **kwargs):
        raise NotImplementedError

    def update(self, model, **kwargs):
        # Some models cannot be updated, so instead of raising a
        # NotImplementedError, we raise a RuntimeError.
        msg = "This model cannot be updated."
        raise RuntimeError(msg)

    def delete(self, pk: str, **kwargs):
        raise NotImplementedError

    def get_waiter(self, name: str) -> Any:
        """
        Get a boto3 waiter object for this service.

        Args:
            name: The name of the waiter to get.

        Returns:
            The boto3 waiter object.

        """
        return self.client.get_waiter(name)


class ReadonlyBoto3ModelManager(Boto3ModelManager):
    def create(self, model, **kwargs):
        msg = "This model cannot be created."
        raise RuntimeError(msg)

    def update(self, model, **kwargs):
        msg = "This model cannot be updated."
        raise RuntimeError(msg)

    def delete(self, pk: str, **kwargs):
        msg = "This model cannot be deleted."
        raise RuntimeError(msg)


class ModelIdentityMixin:
    @property
    def pk(self) -> str | OrderedDict | None:
        """
        Get the primary key of the model instance.

        Returns:
            The primary key of the model instance.

        """
        raise NotImplementedError

    @property
    def arn(self) -> str | None:
        """
        Get the ARN of the model instance.

        Returns:
            The ARN of the model instance.

        """
        msg = "The model does not have an ARN."
        raise ValueError(msg)

    @property
    def name(self) -> str | None:
        """
        Get the name of the model instance.

        Returns:
            The name of the model instance.

        """
        msg = "The model does not have a name."
        raise ValueError(msg)


class classproperty:  # noqa: N801
    """
    This is useful for defining properties that are not instance-specific,
    but rather class-specific.

    Example:
        .. code-block:: python

            class MyClass:
                @classproperty
                def my_property(cls):
                    return "Hello, world!"
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner):
        return self.func(owner)


class ReadonlyPrimaryBoto3Model(  # pylint: disable=abstract-method
    ModelIdentityMixin, ReadonlyBoto3Model
):
    #: The manager for this model
    manager_class: ClassVar[Type[Boto3ModelManager]]

    #: Get the manager for this model, and set it as a class property
    objects: ClassVar[classproperty] = classproperty(lambda cls: cls.manager_class())

    def save(self, **kwargs):
        """
        Save the model.
        """
        msg = "Cannot save a readonly model."
        raise RuntimeError(msg)

    def delete(self):
        """
        Delete the model.
        """
        msg = "Cannot delete a readonly model."
        raise RuntimeError(msg)


class PrimaryBoto3Model(  # pylint: disable=abstract-method
    ModelIdentityMixin, Boto3Model
):
    """
    The base class for all boto3 models that get returned as the primary object
    from a boto3 operation.
    """

    #: The manager for this model
    manager_class: ClassVar[Type[Boto3ModelManager]]

    #: Get the manager for this model, and set it as a class property
    objects: ClassVar[classproperty] = classproperty(lambda cls: cls.manager_class())

    def save(self, **kwargs):
        """
        Save the model.
        """
        if self.pk:
            if hasattr(self.objects, "update"):
                # If the model has a primary key, we assume it is already
                # created and we need to update it.
                # We also assume that the model has a manager.
                return self.objects.update(self, **kwargs)
            msg = f"Model {self.__class__.__name__} is not updatable."
            raise NotUpdatableError(msg)

        return self.objects.create(self, **kwargs)

    def delete(self):
        """
        Delete the model.
        """
        if not self.pk:
            msg = "Cannot delete a model that has not been saved."
            raise ValueError(msg)
        if isinstance(self.pk, OrderedDict):
            return self.objects.delete(**self.pk)
        return self.objects.delete(self.pk)


class PrimaryBoto3ModelQuerySet:
    def __init__(self, results: list[Boto3Model]) -> None:
        """
        Initialize the queryset with a list of results.

        Args:
            results: List of Boto3Model objects
        """
        self.results = results
        if not self.results:
            self.results = []
        self._relationship_cache: Dict[str, Dict[int, Any]] = {}

    def first(self) -> Boto3Model | None:
        """
        Get the first model in the queryset.

        Returns:
            The first Boto3Model object or None if the queryset is empty
        """
        if self.results:
            return self.results[0]
        return None

    def __len__(self) -> int:
        """
        Get the number of models in the queryset.

        Returns:
            The count of Boto3Model objects in the queryset
        """
        return len(self.results)

    def __bool__(self) -> bool:
        """
        Check if the queryset contains any models.

        Returns:
            True if there are models in the queryset, False otherwise
        """
        return bool(self.results)

    def count(self) -> int:
        """
        Get the count of models in the queryset.

        Returns:
            The count of Boto3Model objects in the queryset
        """
        return len(self.results)

    def exists(self) -> bool:
        """
        Check if the queryset contains any models.

        Returns:
            True if there are models in the queryset, False otherwise
        """
        return bool(self.results)

    def order_by(self, field_spec: str) -> "PrimaryBoto3ModelQuerySet":
        """
        Order the queryset by the specified field.

        Args:
            field_spec: Field to order by, can include '__' for nested fields.
                        Prefix with '-' for descending order.

        Returns:
            The ordered queryset
        """
        if not self.results:
            return self

        reverse = False
        if field_spec.startswith("-"):
            reverse = True
            field_spec = field_spec[1:]

        # Use the same field accessor as the filter to get nested fields
        def get_field_value(model):
            filter_obj = Boto3ModelManagerFilter(
                [model], dummy=True, relationship_cache=self._relationship_cache
            )
            try:
                value = filter_obj._get_field_value(model, field_spec)  # noqa: SLF001
            except AttributeError:
                # If the field doesn't exist, return (False, None) so it sorts first
                return (False, None)
            # Handle lists by taking the first value if present
            if isinstance(value, list):
                if value:
                    return (True, value[0])
                # If the list is empty, return (False, None) so it sorts first
                return (False, None)
            if isinstance(value, dict):
                # If it's a dict, we can sort by the first key's value
                keys = list(value.keys())
                if keys:
                    return (True, value[keys[0]])
                # If the dict is empty, return (False, None) so it sorts first
                return (False, None)
            # Return None for missing values so they sort at the beginning
            return (True, value) if value is not None else (False, value)

        self.results = sorted(self.results, key=get_field_value, reverse=reverse)
        return self

    def filter(self, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        """
        Filter the model using the manager's filter method.

        Args:
            **kwargs: The filter criteria.

        Returns:
            A filter object that can be used to filter the model.

        """
        if not self.results:
            msg = "No results to filter"
            raise ValueError(msg)

        # Create a new relationship cache for this filter operation
        self._relationship_cache = {}
        filter_obj = Boto3ModelManagerFilter(
            self.results, relationship_cache=self._relationship_cache, **kwargs
        )
        self.results = filter_obj()
        return self

    def values(self, *fields) -> list[dict]:
        """
        Return a list of dictionaries containing the specified fields.

        Similar to Django's values() method, but returns a list directly instead of a queryset.
        Each dictionary contains the requested fields as keys with their values.

        Args:
            *fields: The field names to include in the output dictionaries.
                     If no fields are specified, all fields will be included.

        Returns:
            A list of dictionaries with the requested fields.
        """
        if not self.results:
            return []

        result_list = []
        for obj in self.results:
            if not fields:
                # If no fields specified, convert the entire object to a dict
                result_list.append(
                    obj.model_dump() if hasattr(obj, "model_dump") else vars(obj)
                )
            else:
                # Extract only the specified fields
                result = {}
                filter_obj = Boto3ModelManagerFilter(
                    [obj], dummy=True, relationship_cache=self._relationship_cache
                )
                for field in fields:
                    try:
                        result[field] = filter_obj._get_field_value(obj, field)  # noqa: SLF001
                    except (AttributeError, KeyError):  # noqa: PERF203
                        result[field] = None
                result_list.append(result)

        return result_list

    def values_list(self, *fields, flat=False) -> list:
        """
        Return a list of tuples containing the values for the specified fields.

        Similar to Django's values_list() method, but returns a list directly instead of a queryset.
        Each tuple contains the values for the requested fields.

        Args:
            *fields: The field names to include in the output tuples.
                     At least one field must be specified.
            flat: If True and only one field is specified, return flat values
                  instead of tuples. Default is False.

        Returns:
            A list of tuples (or flat values if flat=True) with the requested field values.

        Raises:
            ValueError: If flat=True is used with multiple fields or if no fields are specified.
        """
        if not fields:
            msg = "values_list() requires at least one field name"
            raise ValueError(msg)

        if flat and len(fields) > 1:
            msg = "flat=True is only valid when values_list() is called with a single field"
            raise ValueError(msg)

        if not self.results:
            return []

        result_list = []
        for obj in self.results:
            filter_obj = Boto3ModelManagerFilter(
                [obj], dummy=True, relationship_cache=self._relationship_cache
            )
            if flat:
                # Return single values for flat=True
                try:
                    result_list.append(filter_obj._get_field_value(obj, fields[0]))  # noqa: SLF001
                except (AttributeError, KeyError):
                    result_list.append(None)
            else:
                # Return tuples of values
                result = []
                for field in fields:
                    try:
                        result.append(filter_obj._get_field_value(obj, field))  # noqa: SLF001
                    except (AttributeError, KeyError):  # noqa: PERF203
                        result.append(None)
                result_list.append(tuple(result))

        return result_list

    def __iter__(self) -> Iterator[Boto3Model]:
        """
        Enable iteration over the filtered results.

        Returns:
            Iterator over filtered models or values if values()/values_list() was called
        """
        return iter(self.results)

    def all(self) -> list[Boto3Model]:
        """
        Return all models in the queryset.

        Returns:
            A list of Boto3Model objects in the queryset
        """
        return self.results

    def __getitem__(self, index) -> Union["PrimaryBoto3ModelQuerySet", Boto3Model]:
        """
        Enable indexed access to the filtered results.

        Args:
            index: Index or slice to retrieve

        Raises:
            IndexError: If the index is out of range.

        Returns:
            Either a new queryset if a slice is used, or a single model instance
            if a single index is used.
        """
        new_results = self.results[index]
        if not isinstance(new_results, list):
            return new_results
        return self.__class__(new_results)

    def __add__(
        self, other: "PrimaryBoto3ModelQuerySet"
    ) -> "PrimaryBoto3ModelQuerySet":
        """
        Add two querysets together.
        """
        return self.__class__(self.results + other.results)


class Boto3ModelManagerFilter:
    """
    A filter class for Boto3Model objects that provides Django-like filtering capabilities.

    This class allows you to filter a list of :py:class:`Boto3Model` objects
    using familiar Django ORM-style keyword arguments. It supports basic field
    lookups as well as more complex lookups using the ``field__lookup`` syntax.

    Examples:
        # Filter instances where name equals "my-instance"
        filtered = Boto3ModelManagerFilter(instances, name="my-instance")

        # Filter instances where name contains "web" (case-insensitive)
        filtered = Boto3ModelManagerFilter(instances, name__icontains="web")

        # Filter instances where status is "running" AND type is "t2.micro"
        filtered = Boto3ModelManagerFilter(instances, status="running", type="t2.micro")

        # Filter on nested attributes
        filtered = Boto3ModelManagerFilter(instances, tags__Name="web-server")

        # Filter on deeply nested attributes
        filtered = Boto3ModelManagerFilter(instances, network_interfaces__private_ip_address="10.0.0.1")

        # Filter on list field values (automatically searches all items in the list)
        filtered = Boto3ModelManagerFilter(instances, security_groups__group_name="web-sg")

        # Filter on dictionary keys
        filtered = Boto3ModelManagerFilter(instances, metadata__has_key="region")

        # Filter on date/time components
        filtered = Boto3ModelManagerFilter(instances, created_at__date="2023-01-01")
        filtered = Boto3ModelManagerFilter(instances, created_at__year=2023)
        filtered = Boto3ModelManagerFilter(instances, created_at__quarter=2)  # Q2

        # Filter using regular expressions
        filtered = Boto3ModelManagerFilter(instances, name__regex=r"web-[0-9]+")
        filtered = Boto3ModelManagerFilter(instances, name__iregex=r"WEB-[0-9]+")  # Case-insensitive

    Args:
        models: List of Boto3Model objects to filter

    Keyword Args:
        **filters: Django-style keyword arguments for filtering, e.g. name="value", status__exact="running"
    """

    @staticmethod
    def _ensure_utc_datetime(dt_value: Any) -> datetime | None:
        """Convert a value to UTC datetime if it's a datetime with tzinfo."""
        if not isinstance(dt_value, datetime):
            return None

        # Create a copy to avoid modifying the original
        dt_copy = dt_value

        # If it has a timezone, convert to UTC
        if dt_copy.tzinfo is not None:
            return dt_copy.astimezone(ZoneInfo("UTC"))

        # No timezone info, assume it's UTC
        return dt_copy.replace(tzinfo=ZoneInfo("UTC"))

    #: Map of supported lookups to their filter functions
    LOOKUPS: ClassVar[dict[str, Callable[[Any, Any], bool]]] = {
        # String lookups
        "exact": lambda field_val, filter_val: field_val == filter_val,
        "iexact": lambda field_val, filter_val: str(field_val).lower()
        == str(filter_val).lower(),
        "contains": lambda field_val, filter_val: str(filter_val) in str(field_val),
        "icontains": lambda field_val, filter_val: str(filter_val).lower()
        in str(field_val).lower(),
        "startswith": lambda field_val, filter_val: str(field_val).startswith(
            str(filter_val)
        ),
        "istartswith": lambda field_val, filter_val: str(field_val)
        .lower()
        .startswith(str(filter_val).lower()),
        "endswith": lambda field_val, filter_val: str(field_val).endswith(
            str(filter_val)
        ),
        "iendswith": lambda field_val, filter_val: str(field_val)
        .lower()
        .endswith(str(filter_val).lower()),
        # Regular expression lookups
        "regex": lambda field_val, filter_val: re.search(filter_val, str(field_val))
        is not None,
        "iregex": lambda field_val, filter_val: re.search(
            filter_val, str(field_val), re.IGNORECASE
        )
        is not None,
        # Collection lookups
        "in": lambda field_val, filter_val: field_val in filter_val,
        # Comparison lookups
        "gt": lambda field_val, filter_val: field_val > filter_val,
        "gte": lambda field_val, filter_val: field_val >= filter_val,
        "lt": lambda field_val, filter_val: field_val < filter_val,
        "lte": lambda field_val, filter_val: field_val <= filter_val,
        # Null lookups
        "isnull": lambda field_val, filter_val: (field_val is None) == filter_val,
        # Dictionary lookups
        "has_key": lambda field_val, filter_val: isinstance(field_val, dict)
        and filter_val in field_val,
        # Date/time lookups
        "date": lambda field_val, filter_val: (
            isinstance(field_val, datetime)
            and Boto3ModelManagerFilter._ensure_utc_datetime(field_val) is not None
            and Boto3ModelManagerFilter._ensure_utc_datetime(field_val).date()  # type: ignore[union-attr]
            == filter_val
        ),
        "year": lambda field_val, filter_val: (
            isinstance(field_val, datetime)
            and Boto3ModelManagerFilter._ensure_utc_datetime(field_val) is not None
            and Boto3ModelManagerFilter._ensure_utc_datetime(field_val).year  # type: ignore[union-attr]
            == filter_val
        ),
        "iso_year": lambda field_val, filter_val: (
            Boto3ModelManagerFilter._ensure_utc_datetime(field_val) is not None
            and Boto3ModelManagerFilter._ensure_utc_datetime(field_val).isocalendar()[0]  # type: ignore[union-attr]
            == filter_val
        ),
        "month": lambda field_val, filter_val: (
            Boto3ModelManagerFilter._ensure_utc_datetime(field_val) is not None
            and Boto3ModelManagerFilter._ensure_utc_datetime(field_val).month  # type: ignore[union-attr]
            == filter_val
        ),
        "day": lambda field_val, filter_val: (
            Boto3ModelManagerFilter._ensure_utc_datetime(field_val) is not None
            and Boto3ModelManagerFilter._ensure_utc_datetime(field_val).day  # type: ignore[union-attr]
            == filter_val
        ),
        "week": lambda field_val, filter_val: (
            Boto3ModelManagerFilter._ensure_utc_datetime(field_val) is not None
            and Boto3ModelManagerFilter._ensure_utc_datetime(field_val).isocalendar()[1]  # type: ignore[union-attr]
            == filter_val
        ),
        "week_day": lambda field_val, filter_val: (
            Boto3ModelManagerFilter._ensure_utc_datetime(field_val) is not None
            and (Boto3ModelManagerFilter._ensure_utc_datetime(field_val).weekday() + 1)  # type: ignore[union-attr]
            == filter_val
        ),
        "iso_week_day": lambda field_val, filter_val: (
            Boto3ModelManagerFilter._ensure_utc_datetime(field_val) is not None
            and Boto3ModelManagerFilter._ensure_utc_datetime(field_val).isocalendar()[2]  # type: ignore[union-attr]
            == filter_val
        ),
        "quarter": lambda field_val, filter_val: (
            Boto3ModelManagerFilter._ensure_utc_datetime(field_val) is not None
            and (
                (Boto3ModelManagerFilter._ensure_utc_datetime(field_val).month - 1) // 3  # type: ignore[union-attr]
                + 1
            )
            == filter_val
        ),
        "time": lambda field_val, filter_val: (
            Boto3ModelManagerFilter._ensure_utc_datetime(field_val) is not None
            and Boto3ModelManagerFilter._ensure_utc_datetime(field_val).time()  # type: ignore[union-attr]
            == filter_val
        ),
        "hour": lambda field_val, filter_val: (
            Boto3ModelManagerFilter._ensure_utc_datetime(field_val) is not None
            and Boto3ModelManagerFilter._ensure_utc_datetime(field_val).hour  # type: ignore[union-attr]
            == filter_val
        ),
        "minute": lambda field_val, filter_val: (
            Boto3ModelManagerFilter._ensure_utc_datetime(field_val) is not None
            and Boto3ModelManagerFilter._ensure_utc_datetime(field_val).minute  # type: ignore[union-attr]
            == filter_val
        ),
        "second": lambda field_val, filter_val: (
            Boto3ModelManagerFilter._ensure_utc_datetime(field_val) is not None
            and Boto3ModelManagerFilter._ensure_utc_datetime(field_val).second  # type: ignore[union-attr]
            == filter_val
        ),
    }

    def __init__(
        self,
        models: list[Boto3Model],
        relationship_cache: dict[str, dict[int, Any]] | None = None,
        **filters: Any,
    ) -> None:
        """
        Initialize a filter with a list of models and filter criteria.

        Args:
            models: List of Boto3Model objects to filter
            relationship_cache: Optional shared cache for relationship objects
            **filters: Django-style keyword arguments for filtering
        """
        self.models = models
        self.filters = filters
        self._result: list[Boto3Model] | None = None
        self._relationship_cache = relationship_cache or {}
        # Flag to indicate we're just using this instance for field access
        self.dummy = filters.pop("dummy", False)

    def _is_property_method(self, obj: Any, attr_name: str) -> bool:
        """
        Check if the attribute is a property or cached_property method.

        Args:
            obj: The object to check
            attr_name: The attribute name to check

        Returns:
            True if it's a property or cached_property, False otherwise
        """
        if not hasattr(obj.__class__, attr_name):
            return False

        attr = getattr(obj.__class__, attr_name)
        return isinstance(attr, (property, cached_property))

    def _get_property_value(self, obj: Any, attr_name: str) -> Any:
        """
        Get the value of a property method, with caching for the current filter operation.

        Args:
            obj: The object containing the property
            attr_name: The name of the property

        Returns:
            The value of the property
        """
        # Create a cache key based on object id and property name
        cache_key = f"{obj.__class__.__name__}.{attr_name}"
        obj_id = id(obj)

        # Check if we've already cached this property value
        if (
            cache_key in self._relationship_cache
            and obj_id in self._relationship_cache[cache_key]
        ):
            return self._relationship_cache[cache_key][obj_id]

        # If not cached, get the property value
        value = getattr(obj, attr_name)

        # Cache the result
        if cache_key not in self._relationship_cache:
            self._relationship_cache[cache_key] = {}
        self._relationship_cache[cache_key][obj_id] = value

        return value

    def _get_field_value(self, model: Boto3Model, field_name: str) -> Any:  # noqa: PLR0911, PLR0912, PLR0915
        """
        Get a field value from a model, supporting nested attribute access.
        This method handles complex nested structures including lists, dictionaries,
        and relationship properties.

        Args:
            model: The Boto3Model instance
            field_name: The field name, can include various access patterns:
                        - Simple attribute: "name"
                        - Nested attribute: "tags__Name"
                        - Relationship traversal: "cluster__name"
                        - When accessing lists, all items will be searched automatically

        Returns:
            The field value or list of values when traversing through lists

        Raises:
            AttributeError: If the field doesn't exist on the model
        """
        parts = field_name.split("__")
        value = model

        for i, part in enumerate(parts):
            if value is None:
                return None

            # Handle relationship properties
            if i == 0 and self._is_property_method(value, part):
                # Get the relationship property value with caching
                related_obj = self._get_property_value(value, part)

                # If this is the last part, return the related object(s)
                if i == len(parts) - 1:
                    return related_obj

                # If there are more parts, continue traversing
                if related_obj is None:
                    return None

                # Handle both single objects and lists of related objects
                if isinstance(related_obj, list):
                    # Recursively process each related object with the remaining path
                    remaining_path = "__".join(parts[i + 1 :])
                    results = []
                    for item in related_obj:
                        if item is not None:
                            try:
                                result = self._get_field_value(item, remaining_path)
                                if result is not None:
                                    if isinstance(result, list):
                                        results.extend(result)
                                    else:
                                        results.append(result)
                            except (AttributeError, KeyError, TypeError, IndexError):
                                pass
                    return results or None
                # Continue with the next part using the related object
                value = related_obj
                continue

            # Special case for handling dictionary access with nested paths
            if i == 0 and hasattr(value, part) and len(parts) > 1:
                # Get the top-level attribute
                attr_value = getattr(value, part)
                if isinstance(attr_value, dict):
                    # If it's a dictionary, handle the nested parts
                    nested_parts = parts[1:]
                    nested_value = attr_value
                    for nested_part in nested_parts:
                        if nested_value is None or not isinstance(nested_value, dict):
                            return None
                        nested_value = cast(
                            "dict[str, Any]", nested_value.get(nested_part)
                        )
                    return nested_value

            # Handle lists - automatically traverse all items
            if isinstance(value, (list, tuple)) and value:
                # Get the remaining path to check
                remaining_path = "__".join([part, *parts[i + 1 :]])

                # If this is the last part and all items have this attribute
                if i == len(parts) - 1 and all(
                    hasattr(item, part) for item in value if item is not None
                ):
                    return [getattr(item, part) for item in value if item is not None]

                # Get values from all items in the list that have this path
                results = []
                for item in value:
                    if item is not None:
                        try:
                            result = self._get_field_value(item, remaining_path)
                            if result is not None:
                                if isinstance(result, list):
                                    results.extend(result)
                                else:
                                    results.append(result)
                        except (AttributeError, KeyError, TypeError, IndexError):
                            # Skip items that don't have this attribute
                            pass
                return results or None

            # Handle dictionary access
            if isinstance(value, dict):
                value = value.get(part)
            # Handle model attribute access
            elif hasattr(value, part):
                value = getattr(value, part)
            else:
                # Try dictionary-style access for dynamic fields
                try:
                    if hasattr(value, "__getitem__"):
                        value = value[part]
                    else:
                        msg = f"Cannot access {part} on {type(value).__name__}"
                        raise AttributeError(msg)
                except (KeyError, TypeError, IndexError) as e:
                    msg = f"Cannot access {part} on {type(value).__name__}"
                    raise AttributeError(msg) from e

        return value

    def _apply_filter(
        self, model: Boto3Model, field_spec: str, filter_value: Any
    ) -> bool:
        """
        Apply a single filter to a model.

        Args:
            model: The Boto3Model instance
            field_spec: The field specifier (field_name or field_name__lookup)
            filter_value: The value to filter against

        Returns:
            True if the model passes the filter, False otherwise
        """
        # Parse the field specifier into field name and lookup type
        parts = field_spec.split("__")
        field_name = "__".join(parts[0:-1])

        # Default lookup is exact match
        lookup = "exact"
        if len(parts) > 1 and parts[-1] in self.LOOKUPS:
            lookup = parts[-1]
        else:
            # The part after __ might be part of the field path, not a lookup
            field_name = field_spec

        try:
            field_value = self._get_field_value(model, field_name)

            # Handle list values - if any item in the list matches, return True
            if isinstance(field_value, list):
                return any(
                    self.LOOKUPS[lookup](val, filter_value)
                    for val in field_value
                    if val is not None
                )

            # Apply the appropriate lookup filter
            return self.LOOKUPS[lookup](field_value, filter_value)
        except (AttributeError, TypeError):
            # If the field doesn't exist or can't be compared, it doesn't match
            return False

    def all(self) -> list[Boto3Model]:
        """
        Return all models without any filtering.

        Returns:
            List of all Boto3Model objects, after applying any
            :py:class:`Boto3ModelManagerFilter` operations.
        """
        return self()

    def __call__(self) -> list[Boto3Model]:
        """
        Apply all filters and return the filtered list of models.

        Returns:
            List of Boto3Model objects that pass all filters
        """
        if self._result is not None:
            return self._result

        result: list[Boto3Model] = []
        for model in self.models:
            # A model must pass all filters to be included
            if all(
                self._apply_filter(model, field, value)
                for field, value in self.filters.items()
            ):
                result.append(model)  # noqa: PERF401

        self._result = result
        return result

    def __iter__(self) -> Iterator[Boto3Model]:
        """
        Enable iteration over the filtered results.

        Returns:
            Iterator over filtered models
        """
        return iter(self())

    def __len__(self) -> int:
        """
        Get the number of models that passed the filter.

        Returns:
            Number of filtered models
        """
        return len(self())

    def __getitem__(self, index) -> Boto3Model | list[Boto3Model]:
        """
        Enable indexed access to the filtered results.

        Args:
            index: Index or slice to retrieve

        Returns:
            The model at the given index or a slice of models
        """
        return self()[index]

    def __bool__(self) -> bool:
        """
        Check if any models passed the filter.

        Returns:
            True if any models passed the filter, False otherwise
        """
        return bool(self())



import re
from datetime import datetime
from typing import Optional

import pytest
from pydantic import Field

from botocraft.services.abstract import (
    Boto3Model,
    PrimaryBoto3ModelQuerySet,
)


class SimpleTestModel(Boto3Model):
    """A simple model for testing the queryset."""

    name: str
    type: str
    active: bool = True
    score: int = 0
    tags: list[str] = Field(default_factory=list)
    metadata: dict = Field(default_factory=dict)  # Add metadata field for the tests


@pytest.fixture
def empty_queryset():
    """Fixture providing an empty queryset."""
    return PrimaryBoto3ModelQuerySet([])


@pytest.fixture
def test_models():
    """Fixture providing a list of test models."""
    return [
        SimpleTestModel(name="model-1", type="a", score=10, tags=["tag1", "tag2"]),
        SimpleTestModel(name="model-2", type="b", score=20, tags=["tag2", "tag3"]),
        SimpleTestModel(name="model-3", type="a", score=30, tags=["tag3", "tag4"]),
        SimpleTestModel(name="model-4", type="c", score=40, active=False),
    ]


@pytest.fixture
def test_queryset(test_models):
    """Fixture providing a queryset with test models."""
    return PrimaryBoto3ModelQuerySet(test_models)


class TestPrimaryBoto3ModelQuerySet:
    """Test suite for the PrimaryBoto3ModelQuerySet class."""

    def test_init_with_models(self, test_models):
        """Test initializing with models."""
        queryset = PrimaryBoto3ModelQuerySet(test_models)
        assert len(queryset.results) == 4
        assert queryset.results == test_models

    def test_init_empty(self):
        """Test initializing with no models."""
        queryset = PrimaryBoto3ModelQuerySet([])
        assert len(queryset.results) == 0
        assert queryset.results == []

    def test_init_none(self):
        """Test initializing with None."""
        queryset = PrimaryBoto3ModelQuerySet(None)
        assert len(queryset.results) == 0
        assert queryset.results == []

    def test_first(self, test_queryset, test_models):
        """Test getting the first model."""
        assert test_queryset.first() == test_models[0]

    def test_first_empty(self, empty_queryset):
        """Test getting the first model when empty."""
        assert empty_queryset.first() is None

    def test_len(self, test_queryset):
        """Test len() function."""
        assert len(test_queryset) == 4

    def test_len_empty(self, empty_queryset):
        """Test len() function when empty."""
        assert len(empty_queryset) == 0

    def test_bool(self, test_queryset, empty_queryset):
        """Test boolean evaluation."""
        assert bool(test_queryset) is True
        assert bool(empty_queryset) is False

    def test_exists(self, test_queryset, empty_queryset):
        """Test exists() method."""
        assert test_queryset.exists() is True
        assert empty_queryset.exists() is False

    def test_count(self, test_queryset, empty_queryset):
        """Test count() method."""
        assert test_queryset.count() == 4
        assert empty_queryset.count() == 0

    def test_iteration(self, test_queryset, test_models):
        """Test iteration over the queryset."""
        models = []
        for model in test_queryset:
            models.append(model)  # noqa: PERF402
        assert models == test_models

    def test_filter(self, test_queryset):
        """Test filter() method."""
        # Filter by type
        filtered = test_queryset.filter(type="a")
        assert len(filtered) == 2
        assert all(model.type == "a" for model in filtered)

    def test_filter_dict_value(self, test_queryset):
        """Test filtering by a dictionary value."""
        # Add metadata to test models
        test_queryset.results[0].metadata = {
            "environment": "prod",
            "region": "us-west-2",
        }
        test_queryset.results[1].metadata = {
            "environment": "staging",
            "region": "us-west-2",
        }
        test_queryset.results[2].metadata = {
            "environment": "prod",
            "region": "us-east-1",
        }
        test_queryset.results[3].metadata = {"region": "eu-west-1"}

        # Filter by a specific metadata value
        filtered = test_queryset.filter(metadata__environment="prod")
        assert len(filtered) == 2
        assert [m.name for m in filtered] == ["model-1", "model-3"]

    def test_filter_dict_value2(self, test_queryset):
        """Test filtering by a dictionary value."""
        # Add metadata to test models
        test_queryset.results[0].metadata = {
            "environment": "prod",
            "region": "us-west-2",
        }
        test_queryset.results[1].metadata = {
            "environment": "staging",
            "region": "us-west-2",
        }
        test_queryset.results[2].metadata = {
            "environment": "prod",
            "region": "us-east-1",
        }
        test_queryset.results[3].metadata = {"region": "eu-west-1"}

        # Filter by another metadata value
        filtered = test_queryset.filter(metadata__region="us-west-2")
        assert len(filtered) == 2
        assert [m.name for m in filtered] == ["model-1", "model-2"]

    def test_filter_dict_value_combined(self, test_queryset):
        """Test filtering by a dictionary value."""
        # Add metadata to test models
        test_queryset.results[0].metadata = {
            "environment": "prod",
            "region": "us-west-2",
        }
        test_queryset.results[1].metadata = {
            "environment": "staging",
            "region": "us-west-2",
        }
        test_queryset.results[2].metadata = {
            "environment": "prod",
            "region": "us-east-1",
        }
        test_queryset.results[3].metadata = {"region": "eu-west-1"}

        # Filter by combination of metadata values
        filtered = test_queryset.filter(
            metadata__environment="prod", metadata__region="us-east-1"
        )
        assert len(filtered) == 1
        assert filtered[0].name == "model-3"

    def test_filter_dict_value_for_missing_key(self, test_queryset):
        """Test filtering by a dictionary value."""
        # Add metadata to test models
        test_queryset.results[0].metadata = {
            "environment": "prod",
            "region": "us-west-2",
        }
        test_queryset.results[1].metadata = {
            "environment": "staging",
            "region": "us-west-2",
        }
        test_queryset.results[2].metadata = {
            "environment": "prod",
            "region": "us-east-1",
        }
        test_queryset.results[3].metadata = {"region": "eu-west-1"}

        # Filter by missing metadata key
        filtered = test_queryset.filter(metadata__environment="dev")
        assert len(filtered) == 0

    def test_chained_filters(self, test_queryset):
        """Test chaining multiple filters."""
        filtered = test_queryset.filter(active=True).filter(score__gt=15)
        assert len(filtered) == 2
        assert all(model.active and model.score > 15 for model in filtered)

    def test_order_by_simple_field(self, test_queryset):
        """Test ordering by a simple field."""
        ordered = test_queryset.order_by("name")
        assert ordered.first().name == "model-1"  # Alphabetically first
        assert ordered.results[-1].name == "model-4"  # Alphabetically last

    def test_order_by_numeric_field(self, test_queryset):
        """Test ordering by a numeric field."""
        ordered = test_queryset.order_by("score")
        assert [m.score for m in ordered.results] == [10, 20, 30, 40]

        # Test reverse ordering
        ordered = test_queryset.order_by("-score")
        assert [m.score for m in ordered.results] == [40, 30, 20, 10]

    def test_order_by_nested_field(self, test_queryset):
        """Test ordering by a nested field."""
        # Add nested data to test
        test_queryset.results[0].metadata["priority"] = 2
        test_queryset.results[1].metadata["priority"] = 1
        test_queryset.results[2].metadata["priority"] = 3
        # Ensure the last model has a metadata dict too
        test_queryset.results[3].metadata = {}

        ordered = test_queryset.order_by("metadata__priority")
        priorities = [m.metadata.get("priority") for m in ordered.results]
        # None values should come first, then in ascending order
        assert priorities == [None, 1, 2, 3]

    def test_order_by_list_field(self, test_queryset):
        """Test ordering by a field that might return a list."""
        ordered = test_queryset.order_by("tags")
        # Models with tags should be ordered by first tag
        assert all(hasattr(m, "tags") for m in ordered.results)

    def test_chained_filter_and_order(self, test_queryset):
        """Test chaining filter and order_by methods."""
        result = test_queryset.filter(type="a").order_by(
            "score"
        )  # Changed from t2.micro to a
        assert len(result) == 2
        assert [m.score for m in result.results] == [10, 30]

        # Chain in the opposite order
        result = test_queryset.order_by("score").filter(
            type="a"
        )  # Changed from t2.micro to a
        assert len(result) == 2
        assert [m.score for m in result.results] == [10, 30]

    def test_getitem_single_index(self, test_queryset, test_models):
        """Test getting a single item by index."""
        assert test_queryset[0] == test_models[0]
        assert test_queryset[2] == test_models[2]

        # Test negative index
        assert test_queryset[-1] == test_models[-1]

        # Test index out of range
        with pytest.raises(IndexError):
            test_queryset[10]  # Out of range

    def test_getitem_slice(self, test_queryset, test_models):
        """Test getting a slice of items."""
        # Simple slice
        sliced = test_queryset[1:3]
        assert isinstance(sliced, PrimaryBoto3ModelQuerySet)
        assert len(sliced) == 2
        assert sliced.results == test_models[1:3]

        # Slice with step
        sliced = test_queryset[::2]  # Every other item
        assert len(sliced) == 2
        assert sliced.results == test_models[::2]

        # Negative indices in slice
        sliced = test_queryset[-2:]
        assert len(sliced) == 2
        assert sliced.results == test_models[-2:]

        # Empty slice
        sliced = test_queryset[10:20]
        assert len(sliced) == 0

    def test_chained_slice_then_filter(self, test_queryset):
        """Test chaining slice with a filter after it."""
        # Slice then filter
        result = test_queryset[1:3].filter(active=True)
        assert len(result) == 2
        assert all(model.active for model in result)

    def test_chained_filter_then_slice(self, test_queryset):
        """Test chaining filter with a slice after it."""
        result = test_queryset.filter(active=True)[1:2]
        assert len(result) == 1

    def test_chained_slice_then_order_by(self, test_queryset):
        """Test chaining slice with an order_by after it."""
        result = test_queryset[1:4].order_by("score")
        assert len(result) == 3
        assert [m.score for m in result.results] == [20, 30, 40]

    def test_chained_order_by_then_slice(self, test_queryset):
        """Test chaining order_by with a slice after it."""
        result = test_queryset.order_by("score")[1:3]
        assert len(result) == 2
        assert [m.score for m in result.results] == [20, 30]

    def test_chained_slice_all_ops(self, test_queryset):
        """Test chaining slice with both filter and order_by."""
        result = test_queryset.filter(active=True).order_by("score")[1:]
        assert len(result) == 2
        assert [m.score for m in result.results] == [20, 30]


class RelatedModel(Boto3Model):
    """A related model for testing relationship traversal in querysets."""

    id: str
    name: str
    status: str = "active"
    priority: int = 0

    @property
    def child(self) -> Optional["NestedRelatedModel"]:
        """A relationship to a nested model for testing multi-level traversal."""
        if self.id.startswith("has-child"):
            return NestedRelatedModel(
                id=f"child-{self.id}",
                name=f"Child of {self.name}",
                value=self.priority * 10,
            )
        return None


class NestedRelatedModel(Boto3Model):
    """A nested related model for testing deep relationship traversal."""

    id: str
    name: str
    value: int = 0


class ModelWithRelationships(Boto3Model):
    """A model with relationship properties for testing queryset operations."""

    name: str
    type: str
    active: bool = True
    score: int = 0
    related_id: str | None = None
    nested_related_ids: list[str] = Field(default_factory=list)

    @property
    def related(self) -> RelatedModel | None:
        """A single relation property."""
        if not self.related_id:
            return None
        priority = 0
        if "high" in self.related_id:
            priority = 3
        elif "medium" in self.related_id:
            priority = 2
        elif "low" in self.related_id:
            priority = 1
        return RelatedModel(
            id=self.related_id, name=f"Related to {self.name}", priority=priority
        )

    @property
    def nested_relations(self) -> list[RelatedModel]:
        """A one-to-many relation property."""
        return [
            RelatedModel(
                id=rel_id,
                name=f"Nested {idx} of {self.name}",
                priority=int(rel_id[-1] if rel_id.startswith("has-child") else 0),
            )
            for idx, rel_id in enumerate(self.nested_related_ids)
        ]


@pytest.fixture
def models_with_relationships():
    """Fixture providing test models with relationships."""
    return [
        ModelWithRelationships(
            name="model-1",
            type="a",
            score=10,
            related_id="low-priority",
            nested_related_ids=["nested-1", "has-child-1"],
        ),
        ModelWithRelationships(
            name="model-2",
            type="b",
            score=20,
            related_id="medium-priority",
            nested_related_ids=["nested-2"],
        ),
        ModelWithRelationships(
            name="model-3",
            type="a",
            score=30,
            related_id="high-priority",
            nested_related_ids=["has-child-2"],
        ),
        ModelWithRelationships(
            name="model-4", type="c", score=40, active=False, related_id=None
        ),
    ]


@pytest.fixture
def relationship_queryset(models_with_relationships):
    """Fixture providing a queryset with models having relationships."""
    return PrimaryBoto3ModelQuerySet(models_with_relationships)


class TestRelationshipTraversal:
    """Test suite for relationship traversal in PrimaryBoto3ModelQuerySet."""

    def test_filter_by_relationship_field(self, relationship_queryset):
        """Test filtering by a field on a related object."""
        filtered = relationship_queryset.filter(related__name__contains="model-1")
        assert len(filtered) == 1
        assert filtered.first().name == "model-1"

    def test_filter_by_relationship_priority(self, relationship_queryset):
        """Test filtering by a numeric field on related objects."""
        # Find models with high priority related objects
        filtered = relationship_queryset.filter(related__priority__gt=2)
        assert len(filtered) == 1
        assert all(f.related.priority > 2 for f in filtered)

    def test_filter_by_relationship_priority_is_null(self, relationship_queryset):
        """Test filtering where related priority is None."""
        filtered = relationship_queryset.filter(related__priority__isnull=False)
        assert len(filtered) == 3

    def test_order_by_relationship_field(self, relationship_queryset):
        """Test ordering by a field on a related object."""
        # Order by priority on the related object
        ordered = relationship_queryset.order_by("related__priority")

        # Null values come first, then in ascending order of priority
        models = list(ordered)
        assert models[0].name == "model-4"  # No related (None)
        assert models[1].name == "model-1"  # Low priority
        assert models[2].name == "model-2"  # Medium priority
        assert models[3].name == "model-3"  # Low priority

    def test_order_by_relationship_field_reverse(self, relationship_queryset):
        """Test ordering by a field on a related object in reverse."""
        ordered = relationship_queryset.order_by("-related__priority")
        models = list(ordered)
        assert models[0].name == "model-3"  # High priority
        assert models[3].name == "model-4"  # No related (None)

    def test_filter_on_many_relationship(self, relationship_queryset):
        """Test filtering on a one-to-many relationship."""
        # Find models that have any nested relation with "Nested 0" in the name
        filtered = relationship_queryset.filter(
            nested_relations__name__contains="Nested 0"
        )
        assert len(filtered) == 3  # Models 1, 2, and 3

    def test_filter_on_nested_relationship(self, relationship_queryset):
        """Test filtering on a nested relationship."""
        filtered = relationship_queryset.filter(
            nested_relations__id__startswith="has-child"
        )
        assert len(filtered) == 2
        assert sorted([m.name for m in filtered]) == ["model-1", "model-3"]

    def test_multi_level_traversal(self, relationship_queryset):
        """Test filtering through multiple levels of relationships."""
        # Find models that have a nested relation with a child whose name contains "Child"
        filtered = relationship_queryset.filter(
            nested_relations__child__name__contains="Child"
        )
        assert len(filtered) == 2
        assert sorted([m.name for m in filtered]) == ["model-1", "model-3"]

    def test_multi_level_traversal_order_by(
        self, relationship_queryset, models_with_relationships
    ):
        # Test ordering by a deeply nested field
        ordered = relationship_queryset.order_by("nested_relations__child__value")

        # Find the last model - should be the one with highest child value
        models_with_child = [
            m
            for m in models_with_relationships
            if any(rel_id.startswith("has-child") for rel_id in m.nested_related_ids)
        ]
        highest_priority = max(
            m.related.priority if m.related else 0 for m in models_with_child
        )
        assert ordered[-1].related.priority == highest_priority

    def test_chained_filtering_with_relationships(self, relationship_queryset):
        """Test chaining multiple filters involving relationships."""
        # Filter by type and then by related field
        result = relationship_queryset.filter(type="a").filter(related__priority__gt=0)
        assert len(result) == 2
        assert all(
            m.type == "a" and m.related and m.related.priority > 0 for m in result
        )

        # Filter in reverse order
        result = relationship_queryset.filter(related__priority__gt=0).filter(type="a")
        assert len(result) == 2
        assert all(
            m.type == "a" and m.related and m.related.priority > 0 for m in result
        )

    def test_filter_and_order_with_relationships(self, relationship_queryset):
        """Test filtering and ordering with relationships."""
        # Filter models with type "a" and order by related priority
        result = relationship_queryset.filter(type="a").order_by("related__priority")
        assert len(result) == 2
        assert [m.related.priority for m in result.results] == [1, 3]


class TestValuesAndValuesList:
    """Test suite for values() and values_list() methods."""

    @pytest.fixture
    def test_models(self) -> list[Boto3Model]:
        """Create a list of test models."""

        class SimpleModel(Boto3Model):
            id: str
            name: str
            age: int
            active: bool
            created_at: datetime = Field(default_factory=datetime.now)
            tags: dict = Field(default_factory=dict)

            @property
            def full_name(self):
                return f"Full: {self.name}"

            @property
            def nested_data(self):
                return {"key1": "value1", "key2": self.age * 2}

        return [
            SimpleModel(
                id="1", name="Alice", age=30, active=True, tags={"role": "admin"}
            ),
            SimpleModel(
                id="2", name="Bob", age=25, active=False, tags={"role": "user"}
            ),
            SimpleModel(
                id="3", name="Charlie", age=35, active=True, tags={"role": "user"}
            ),
        ]

    def test_values_no_fields(self, test_models):
        """Test values() with no specified fields returns all fields."""
        qs = PrimaryBoto3ModelQuerySet(test_models)
        values = qs.values()  # No need to call list() anymore

        assert len(values) == 3
        assert isinstance(values[0], dict)
        assert "id" in values[0]
        assert "name" in values[0]
        assert "age" in values[0]
        assert "active" in values[0]
        assert values[0]["id"] == "1"
        assert values[1]["name"] == "Bob"
        assert values[2]["age"] == 35

    def test_values_specific_fields(self, test_models):
        """Test values() with specific fields returns only those fields."""
        qs = PrimaryBoto3ModelQuerySet(test_models)
        values = qs.values("id", "name")  # No need to call list() anymore

        assert len(values) == 3
        assert set(values[0].keys()) == {"id", "name"}
        assert values[0]["id"] == "1"
        assert values[0]["name"] == "Alice"
        assert "age" not in values[0]

    def test_values_with_property(self, test_models):
        """Test values() with a property field."""
        qs = PrimaryBoto3ModelQuerySet(test_models)
        values = list(qs.values("id", "full_name"))

        assert len(values) == 3
        assert values[0]["full_name"] == "Full: Alice"
        assert values[1]["full_name"] == "Full: Bob"

    def test_values_with_nested_property(self, test_models):
        """Test values() with a nested property field."""
        qs = PrimaryBoto3ModelQuerySet(test_models)
        values = list(qs.values("id", "nested_data__key2"))

        assert len(values) == 3
        assert values[0]["nested_data__key2"] == 60  # 30 * 2
        assert values[1]["nested_data__key2"] == 50  # 25 * 2

    def test_values_with_dict_field(self, test_models):
        """Test values() with a dictionary field."""
        qs = PrimaryBoto3ModelQuerySet(test_models)
        values = list(qs.values("id", "tags__role"))

        assert len(values) == 3
        assert values[0]["tags__role"] == "admin"
        assert values[1]["tags__role"] == "user"

    def test_values_with_nonexistent_field(self, test_models):
        """Test values() with a field that doesn't exist."""
        qs = PrimaryBoto3ModelQuerySet(test_models)
        values = list(qs.values("id", "nonexistent"))

        assert len(values) == 3
        assert values[0]["id"] == "1"
        assert values[0]["nonexistent"] is None

    def test_values_empty_queryset(self):
        """Test values() with an empty queryset."""
        qs = PrimaryBoto3ModelQuerySet([])
        values = list(qs.values("id", "name"))

        assert len(values) == 0

    def test_values_list_single_field(self, test_models):
        """Test values_list() with a single field."""
        qs = PrimaryBoto3ModelQuerySet(test_models)
        values = list(qs.values_list("name"))

        assert len(values) == 3
        assert values[0] == ("Alice",)
        assert values[1] == ("Bob",)
        assert values[2] == ("Charlie",)

    def test_values_list_multiple_fields(self, test_models):
        """Test values_list() with multiple fields."""
        qs = PrimaryBoto3ModelQuerySet(test_models)
        values = list(qs.values_list("id", "name", "age"))

        assert len(values) == 3
        assert values[0] == ("1", "Alice", 30)
        assert values[1] == ("2", "Bob", 25)
        assert values[2] == ("3", "Charlie", 35)

    def test_values_list_flat(self, test_models):
        """Test values_list() with flat=True."""
        qs = PrimaryBoto3ModelQuerySet(test_models)
        values = list(qs.values_list("name", flat=True))

        assert len(values) == 3
        assert values[0] == "Alice"
        assert values[1] == "Bob"
        assert values[2] == "Charlie"

    def test_values_list_with_property(self, test_models):
        """Test values_list() with a property field."""
        qs = PrimaryBoto3ModelQuerySet(test_models)
        values = list(qs.values_list("full_name", flat=True))

        assert len(values) == 3
        assert values[0] == "Full: Alice"
        assert values[1] == "Full: Bob"

    def test_values_list_with_nested_field(self, test_models):
        """Test values_list() with a nested field."""
        qs = PrimaryBoto3ModelQuerySet(test_models)
        values = list(qs.values_list("id", "tags__role"))

        assert len(values) == 3
        assert values[0] == ("1", "admin")
        assert values[1] == ("2", "user")

    def test_values_list_flat_with_multiple_fields_raises_error(self, test_models):
        """Test values_list() with flat=True and multiple fields raises ValueError."""
        qs = PrimaryBoto3ModelQuerySet(test_models)

        with pytest.raises(
            ValueError,
            match=re.escape(
                "flat=True is only valid when values_list() is called with a single field"
            ),
        ):
            list(qs.values_list("id", "name", flat=True))

    def test_values_list_no_fields_raises_error(self, test_models):
        """Test values_list() with no fields raises ValueError."""
        qs = PrimaryBoto3ModelQuerySet(test_models)

        with pytest.raises(
            ValueError,
            match=re.escape("values_list() requires at least one field name"),
        ):
            list(qs.values_list())

    def test_values_list_nonexistent_field(self, test_models):
        """Test values_list() with a field that doesn't exist."""
        qs = PrimaryBoto3ModelQuerySet(test_models)
        values = list(qs.values_list("id", "nonexistent"))

        assert len(values) == 3
        assert values[0] == ("1", None)
        assert values[1] == ("2", None)

    def test_chaining_filter_and_values(self, test_models):
        """Test chaining filter() and values()."""
        qs = PrimaryBoto3ModelQuerySet(test_models)
        values = qs.filter(active=True).values(
            "id", "name"
        )  # No need to call list() anymore

        assert len(values) == 2
        assert values[0]["id"] == "1"
        assert values[0]["name"] == "Alice"
        assert values[1]["id"] == "3"
        assert values[1]["name"] == "Charlie"

    def test_chaining_filter_and_values_list(self, test_models):
        """Test chaining filter() and values_list()."""
        qs = PrimaryBoto3ModelQuerySet(test_models)
        values = qs.filter(age__gt=25).values_list(
            "name", flat=True
        )  # No need to call list() anymore

        assert len(values) == 2
        assert values[0] == "Alice"
        assert values[1] == "Charlie"
